from torchvision import models
from torchvision.models import resnet34
from .dataset import (Dataset, PickingDataset, ClassifierDataset1D,
                      spectrogram)
from torch.utils.data import DataLoader
import torch
from torch import nn
import numpy as np
from loguru import logger
from tqdm import tqdm
import pickle
from sklearn.metrics import confusion_matrix
# from plotcm import plot_confusion_matrix
from uquake.core import Stream, Trace, UTCDateTime
# from torchaudio.transforms import MelSpectrogram, AmplitudeToDB
from PIL import Image, ImageOps
from torch import nn
from .resnet1d import ResNet1D
from importlib import reload
from ipdb import set_trace
# import urllib.request
import requests
from io import BytesIO
from uquake.core.util.requests import download_file_from_url

from .params import *


# class SeismicResnet34(type(resnet34())):
#     def __init__(self, num_input_channels=3, n_out_features=3):
#         super(SeismicResnet34, self).__init__()
#         self.conv1 = nn.Conv2d(num_input_channels, 64, kernel_size=7, stride=2,
#                                padding=3, bias=False)
#
#         self.fc = nn.Linear(self.fc.in_features, n_out_features)


class EventClassifier(object):

    model_url = \
            'https://www.dropbox.com/s/5d76v8hfi3dwsm0/classification_model.useis?dl=1'
    n_out_features = 3
    label_mapping = {
        'noise': np.array([1., 0., 0.]),
        'blast': np.array([0., 1., 0.]),
        'seismic event': np.array([0., 0., 1.])
    }

    def __init__(self, gpu: bool = True, learning_rate: float = 0.001, model_id=None,
                 model=models.resnet34(), weight_decay: float = 0.001):

        # define the model
        self.model = model
        num_input_channels = 3
        self.num_imput_channels = num_input_channels
        self.model.conv1 = nn.Conv2d(num_input_channels, 64, kernel_size=7, stride=2,
                                     padding=3, bias=False)

        self.model.fc = nn.Linear(self.model.fc.in_features, self.n_out_features)

        self.model_id = model_id

        self.n_features = self.n_out_features
        self.num_classes = self.n_out_features
        self.learning_rate = learning_rate
        self.weight_decay = weight_decay

        # Replace the last fully-connected layer to output the desired number of classes
        self.model.fc = nn.Linear(self.model.fc.in_features, self.n_out_features)

        # self.optimizer = torch.optim.SGD(self.model.parameters(), lr=learning_rate)
        self.optimizer = torch.optim.Adam(self.model.parameters(), lr=learning_rate,
                                          weight_decay=weight_decay)
        self.criterion = nn.CrossEntropyLoss()
        self.gpu = gpu

        self.device = self.select_device(gpu=gpu)

        self.model = self.model.to(self.device)
        self.accuracies = []
        self.losses = []

        self.iteration = 0
        # Download the model from Dropbox

    @classmethod
    def from_pretrained_model(cls, model, gpu: bool = True):
        cls_tmp = cls(gpu=gpu, model=model)
        # cls_tmp.model = model.to(cls_tmp.device)
        return cls_tmp

    @staticmethod
    def select_device(gpu=True):
        if gpu:
            device = torch.device("cuda:0" if
                                  torch.cuda.is_available() else "cpu")
            if device == "cpu":
                logger.warning('GPU is not available, the CPU will be used')
            else:
                logger.info('GPU will be used')
        else:
            device = torch.device('cpu')
            logger.info('The CPU will be used')

        return device


    @classmethod
    def from_pretrained_model_file(cls, file_name, gpu: bool = True):
        # with open(path, 'rb') as input_file:
        model = torch.load(file_name, map_location=cls.select_device(gpu=gpu))
        return cls.from_pretrained_model(model, gpu=gpu)

    # @classmethod
    # def load_model(cls, file_name):
    #     return pickle.load(open('filename', 'rb'))

    def train(self, dataset: Dataset, batch_size: int):

        self.model.train()

        self.accuracies = []
        self.losses = []

        training_loader = DataLoader(dataset, batch_size=batch_size,
                                     shuffle=True)

        for inputs, targets in tqdm(training_loader):
            self.iterate(inputs, targets)
            torch.cuda.empty_cache()

    def forward(self, x):
        return self.model.forward(x)

    def iterate(self, inputs, targets):

        self.optimizer.zero_grad()

        inputs = inputs.to(self.device)
        targets = targets.to(self.device)

        # with torch.no_grad():
        predictions = self.model(inputs)

        loss = self.criterion(predictions, targets)
        # loss.requires_grad = True
        loss.backward()
        self.optimizer.step()

        target_indices = torch.argmax(targets, dim=1)
        predicted_indices = torch.argmax(predictions, dim=1)
        loss = self.criterion(predictions, target_indices)
        accuracy = (target_indices == predicted_indices).sum() / len(target_indices)
        self.accuracies.append(accuracy.item())
        self.losses.append(loss.item())
        self.iteration += 1

    def validate(self, dataset: Dataset, batch_size: int):
        self.model.eval()
        training_loader = DataLoader(dataset, batch_size=batch_size,
                                     shuffle=True)
        self.validation_targets = np.zeros(len(dataset))
        self.validation_predictions = np.zeros(len(dataset))
        accuracy = []
        loss = []
        with torch.no_grad():
            for i, (inputs, targets) in enumerate(tqdm(training_loader)):
                # inputs = inputs.view(inputs.size()[0], -1, inputs.size()[1],
                #                      inputs.size()[2])
                inputs = inputs.to(self.device)

                start = batch_size * i
                end = start + batch_size
                self.validation_targets[start:end] = torch.argmax(targets).cpu().numpy()
                targets = targets.to(self.device)

                # with torch.no_grad() pytorch does not compute the gradient,
                # the gradient takes a significant amount of memory on the gpu.
                # using torch.no_grad() consequently allows significantly
                # increasing the batch size
                # with torch.no_grad():
                outputs = self.model(inputs)

                self.validation_predictions[start:end] = torch.argmax(
                    outputs).cpu().numpy()

                target_indices = torch.argmax(targets, dim=1)
                predicted_indices = torch.argmax(outputs, dim=1)
                tmp = (target_indices == predicted_indices).sum() / len(target_indices)
                accuracy.append(float(tmp.cpu().numpy()))
                loss.append(float(self.criterion(outputs, target_indices).cpu().numpy()))
        self.model.train()
        return accuracy, loss

    def display_memory(self):
        torch.cuda.empty_cache()
        memory = torch.cuda.memory_allocated(self.device)
        logger.info("{:.3f} GB".format(memory / 1024 ** 3))

    # @property
    # def model_state(self):
    #     self.model.to('cpu')
    #     out_dict = {'model_state': self.model.state_dict(),
    #                 'out_features': self.n_features,
    #                 'label_map': self.label_mapping,
    #                 'learning_rate': self.learning_rate,
    #                 'model_id': self.model_id}
    #     return out_dict

    def save(self, file_name):
        return torch.save(self.model.state_dict(), file_name)

    def write(self, file_name):
        return self.save(file_name)

    @classmethod
    def from_model_state(cls, state_dict, gpu=True):
        ec = cls(gpu=gpu)
        ec.model.load_state_dict(state_dict)
        ec.model.eval()
        ec.model.to(cls.select_device(gpu=gpu))
        return ec

    @classmethod
    def read(cls, file_name, gpu=True):
        device = cls.select_device(gpu=gpu)
        state_dict = torch.load(file_name, map_location=device)
        # model_state = pickle.load(open(file_name, 'rb'))
        return cls.from_model_state(state_dict, gpu=gpu)

    @classmethod
    def load(cls, gpu=True):
        """
        load the most recent model
        """

        model_data = download_file_from_url(cls.model_url)

        # Load the model from the BytesIO object

        device = cls.select_device(gpu=gpu)

        state_dict = torch.load(model_data, map_location=device)
        return cls.from_model_state(state_dict, gpu=gpu)
        # model = MyModel(...)
        # model.model.load_state_dict(state_dict['model_state'])
        # model.n_features = state_dict['out_features']
        # model.label_mapping = state_dict['label_map']
        # model.learning_rate = state_dict['learning_rate']
        # model.model_id = state_dict['model_id']

    @staticmethod
    def measure_accuracy(targets, predictions):
        max_index = predictions.max(dim=1)[1]
        return (max_index == targets).sum() / len(targets)


class EventClassifier2(EventClassifier):
    model_url = 'https://www.dropbox.com/s/klk2nfalqt8ugjj/1s_model_2023_04_30.pt?dl=1'

    def __init__(self, gpu: bool = True,
                 learning_rate: float = 0.001, model=models.resnet34(), model_id=None,
                 weight_decay: float = 0.001, dropout_prob: float = 0.3):
        super().__init__(gpu=gpu,
                         learning_rate=learning_rate, model_id=model_id, model=model,
                         weight_decay=weight_decay)
        self.num_input_channels = 1

        self.model.conv1 = nn.Conv2d(self.num_input_channels, 64, kernel_size=7,
                                     stride=2, padding=3, bias=False)

        self.model.to(self.device)

        self.dropout = nn.Dropout(p=dropout_prob)

    def forward(self, x):
        x = self.resnet34.conv1(x)
        x = self.resnet34.bn1(x)
        x = self.resnet34.relu(x)
        x = self.resnet34.maxpool(x)

        # Stage 1
        x = self.resnet34.layer1(x)

        # Stage 2
        x = self.resnet34.layer2(x)

        # Stage 3
        x = self.resnet34.layer3(x)

        # Stage 4
        x = self.resnet34.layer4(x)

        # Apply dropout to the output of the last convolutional layer
        x = self.dropout(x)

        x = self.resnet34.avgpool(x)
        x = x.view(x.size(0), -1)
        x = self.resnet34.fc(x)
        return x

    @staticmethod
    def trace2spectrogram(trace: Trace):
        return generate_spectrogram(Stream(traces=[trace]))

    @staticmethod
    def stream2spectrogram(stream: Stream):
        return generate_spectrogram(stream)

    @property
    def model_state(self):
        out_dict = {'model_state': self.model.state_dict(),
                    'out_features': self.n_features,
                    'label_map': self.label_mapping,
                    'learning_rate': self.learning_rate,
                    'model_id': self.model_id}
        return out_dict

    # def save(self, file_name):
    #     return pickle.dump(self.model_state, open(file_name, 'wb'))
    #
    # def write(self, file_name):
    #     return self.save(file_name)
    #
    # @classmethod
    # def from_model_state(cls, model_state, gpu=True):
    #     ec = cls(model_state['out_features'], model_state['label_map'],
    #              gpu=gpu, learning_rate=model_state['learning_rate'],
    #              model_id=model_state['model_id'])
    #     ec.model.load_state_dict(model_state['model_state'])
    #     ec.model.eval()
    #     return ec

    # @classmethod
    # def read(cls, file_name, gpu=True):
    #     model_state = pickle.load(open(file_name, 'rb'))
    #     return cls.from_model_state(model_state, gpu=gpu)


def read_classifier(file_name):
    return EventClassifier.read(file_name)


def generate_spectrogram(stream: Stream):

    specs = []
    trs = []
    for tr in stream.copy():
        # check if tr contains NaN
        if np.any(np.isnan(tr.data)):
            continue
        trs.append(tr.copy())

    st2 = Stream(traces=trs)
    st2 = st2.detrend('demean').detrend('linear')

    st2 = st2.taper(max_percentage=1, max_length=1e-2)

    for tr in st2:
        spec = spectrogram(tr)
        spec /= np.max(spec)
        spec = spec / np.max(spec) * 255
        spec = Image.fromarray(np.array(spec.tolist()).astype(np.uint8))
        specs.append(spec)

    return specs


class EventClassifier1D(EventClassifier):
    def __init__(self, n_classes: int, in_channels: int = 1,
                 base_filters: int = 16, kernel_size: int = 7,
                 stride: int = 3, groups: int = 1, n_block: int = 8,
                 learning_rate=1e-3, gpu: bool = True, weight_decay: float = 0.001):

        if gpu:
            device = torch.device("cuda:0" if
                                  torch.cuda.is_available() else "cpu")
            if device == "cpu":
                logger.warning('GPU is not available, the CPU will be used')
            else:
                logger.info('GPU will be used')
        else:
            device = 'cpu'
            logger.info('The CPU will be used')

        self.device = device

        self.model = ResNet1D(in_channels, base_filters, kernel_size,
                              stride, groups, n_block,
                              n_classes).to(self.device)

        # self.model = Conv1DModel()

        self.in_channels = in_channels
        self.base_filters = base_filters
        self.kernel_size = kernel_size
        self.stride = stride
        self.n_classes = n_classes
        self.groups = groups
        self.n_block = n_block
        self.gpu = gpu

        self.model = ResNet1D(in_channels, base_filters, kernel_size,
                              stride, groups, n_block,
                              n_classes).to(self.device)

        self.optimizer = torch.optim.Adam(self.model.parameters(),
                                          lr=learning_rate,
                                          weight_decay=weight_decay)
        # self.optimizer = torch.optim.SGD(self.model.parameters(),
        #                                  lr=learning_rate)
        self.criterion = nn.CrossEntropyLoss()

        # self.display_memory()
        self.losses = []

        self.validation_predictions = None
        self.validation_targets = None

    def train(self, dataset: ClassifierDataset1D, batch_size: int):

        training_loader = DataLoader(dataset, batch_size=batch_size,
                                     shuffle=True)

        for inputs, targets in tqdm(training_loader):
            inputs = inputs.view(inputs.size()[0], -1, inputs.size()[1])
            self.iterate(inputs, targets)
            torch.cuda.empty_cache()

    def iterate(self, inputs, targets):

        self.optimizer.zero_grad()

        inputs = inputs.to(self.device)
        targets = targets.to(self.device)

        predictions = self.model(inputs)

        # loss = self.criterion(predictions, targets.view(len(targets), -1))
        loss = self.criterion(predictions, targets)
        # print(loss.item())
        loss.backward()
        self.optimizer.step()

        # print(loss.item())
        self.losses.append(loss.cpu().item())
        self.iteration += 1

    def validate(self, dataset: ClassifierDataset1D, batch_size: int):
        self.model.eval()
        training_loader = DataLoader(dataset, batch_size=batch_size,
                                     shuffle=True)
        self.validation_targets = np.zeros(len(dataset))
        self.validation_predictions = np.zeros(len(dataset))
        accuracy = 0
        for i, (inputs, targets) in enumerate(tqdm(training_loader)):
            inputs = inputs.view(inputs.size()[0], -1, inputs.size()[1])
            inputs = inputs.to(self.device)

            start = batch_size * i
            end = start + batch_size
            self.validation_targets[start:end] = targets
            targets = targets.to(self.device)

            # with torch.no_grad() pytorch does not compute the gradient,
            # the gradient takes a significant amount of memory on the gpu.
            # using torch.no_grad() consequently allows significantly
            # increasing the batch size
            with torch.no_grad():
                outputs = self.model(inputs)

            self.validation_predictions[start:end] = outputs.max(dim=1)[1].cpu(
            ).detach().numpy()

            accuracy += self.measure_accuracy(targets, outputs)
        self.model.train()
        return accuracy.cpu().detach().numpy() / (i + 1)


class Conv1DModel(nn.Module):
    def __init__(self):
        super(Conv1DModel, self).__init__()
        self.conv_layers = nn.Sequential(
            nn.Conv1d(1, 64, 15),
            nn.ReLU(),
            nn.Conv1d(64, 64, 15),
            nn.ReLU(),
            nn.Conv1d(64, 64, 15),
            nn.ReLU(),
            nn.Conv1d(64, 128, 15),
            nn.ReLU(),
            nn.Conv1d(128, 128, 15),
            nn.ReLU(),
            nn.Conv1d(128, 128, 15),
            nn.ReLU()
        )

        self.dense_layers = nn.Sequential(
            nn.Dropout(0.2),
            nn.Linear(in_features=128 * 172, out_features=256),
            nn.ReLU(),
            nn.Linear(in_features=256, out_features=256),
            nn.ReLU(),
            nn.Linear(in_features=256, out_features=256),
            nn.ReLU(),
            nn.Linear(in_features=256, out_features=1))

    def forward(self, X):
        out = self.conv_layers(X)
        out = out.view(out.size(0), -1)
        out = self.dense_layers(out)
        # input(out)
        return out


# class AIPicker(EventClassifier):
class AIPicker(object):
    def __init__(self, in_channels: int = 1, base_filters: int = 64,
                 kernel_size: int = 15, stride: int = 3, n_classes: int = 1,
                 groups: int = 1, n_block: int = 16, learning_rate=1e-5,
                 gpu: bool = True, n_sample: int = 256,
                 sampling_rate: int=6000, to_device=True):
        self.in_channels = in_channels
        self.base_filters = base_filters
        self.kernel_size = kernel_size
        self.stride = stride
        self.n_classes = n_classes
        self.groups = groups
        self.n_block = n_block
        self.gpu = gpu
        self.n_sample = n_sample
        self.sampling_rate = sampling_rate

        if gpu:
            device = torch.device("cuda:0" if
                                  torch.cuda.is_available() else "cpu")
            if device == "cpu":
                logger.warning('GPU is not available, the CPU will be used')
            else:
                logger.info('GPU will be used')
        else:
            device = 'cpu'
            logger.info('The CPU will be used')

        self.device = device

        self.model = self.init_model()

        if to_device:
            self.model.to(self.device)
        # self.model = ResNet1D(self.in_channels, self.base_filters, self.kernel_size,
        #                       self.stride, self.groups, self.n_block,
        #                       n_classes).to(self.device)

        self.optimizer = torch.optim.Adam(self.model.parameters(),
                                          lr=learning_rate)
        self.criterion = nn.MSELoss()

        self.losses = []

        self.validation_predictions = None
        self.validation_targets = None

    def init_model(self):
        return ResNet1D(self.in_channels, self.base_filters, self.kernel_size,
                        self.stride, self.groups, self.n_block,
                        self.n_classes)

    def train(self, dataset: PickingDataset, batch_size: int):

        self.model.train()

        training_loader = DataLoader(dataset, batch_size=batch_size,
                                     shuffle=True)

        for inputs, targets in tqdm(training_loader):
            inputs = inputs.view(inputs.size()[0], -1, inputs.size()[1])
            self.iterate(inputs, targets)
            torch.cuda.empty_cache()

    def iterate(self, inputs, targets):

        self.optimizer.zero_grad()

        inputs = inputs.to(self.device)
        targets = targets.to(self.device)
        predictions = self.model(inputs)

        loss = self.criterion(predictions, targets.view(len(targets), -1))
        # print(loss.item())
        loss.backward()
        self.optimizer.step()

        # print(loss.item())
        self.losses.append(loss.cpu().item())

    def predict(self, inputs):
        if isinstance(inputs, np.ndarray):
            if len(inputs.shape) == 1:
                dim1 = 1
                dim2 = inputs.shape[0]
            else:
                dim1 = inputs.shape[0]
                dim2 = inputs.shape[1]

            inputs = torch.from_numpy(inputs)
        elif not isinstance(inputs, torch.Tensor):
            raise TypeError

        inputs = inputs.view(dim1, -1, dim2).to(self.device)
        with torch.no_grad():
            outputs_tensor = self.model(inputs)

        outputs = [out.item() for out in outputs_tensor.cpu()]
        return outputs

    def predict_trace(self, trace: Trace, pick_time: UTCDateTime, phase: str):
        # import matplotlib.pyplot as plt

        trace = trace.resample(sampling_rate=int(self.sampling_rate))

        n_sample = int((pick_time - trace.stats.starttime)
                       * trace.stats.sampling_rate)

        if phase == 'P':
            n_sample_start = int(n_sample - self.n_sample * 0.5)
        else:
            n_sample_start = int(n_sample - self.n_sample * 0.5)

        n_sample_end = int(n_sample_start + self.n_sample)

        data = trace.data[n_sample_start: n_sample_end]
        data -= np.mean(data)
        data /= np.max(np.abs(data))

        predicted_arrival = self.predict(data)[0]

        # import matplotlib.pyplot as plt
        # plt.clf()
        # plt.plot(data)
        # plt.axvline(predicted_arrival)
        # plt.show()
        # input(predicted_arrival)

        # return n_sample_start + predicted_arrival * self.n_sample

        new_pick_time = trace.stats.starttime + \
                        (n_sample_start + predicted_arrival *
                        self.n_sample) / self.sampling_rate

        return new_pick_time

    def predict_stream(self, st: Stream, pick_times):
        pass

    def validate(self, dataset: PickingDataset, batch_size):
        predictions = []
        training_loader = DataLoader(dataset, batch_size=batch_size,
                                     shuffle=True)

        losses = []
        for inputs, targets in tqdm(training_loader):
            inputs = inputs.view(inputs.size()[0], -1, inputs.size()[1])

            inputs = inputs.to(self.device)
            targets = targets.to(self.device)
            with torch.no_grad():
                ps = self.model(inputs)

            for p in ps.cpu():
                predictions.append(p.item())

            loss = self.criterion(ps, targets.view(len(targets), -1))
            # print(loss.cpu().item())
            losses.append(loss.cpu().item())
        return predictions, np.mean(losses), np.std(losses)

    def save(self, file_name):
        with(open(file_name, 'wb')) as f_out:
            pickle.dump(self, f_out)
        # torch.save(self.model.state_dict(), file_name)

    def write(self, file_name):
        self.save(file_name)

    @classmethod
    def read(cls, file_name):
        with open(file_name, 'rb') as f_in:
            return pickle.load(f_in)

    @classmethod
    def from_model(cls, file_name, gpu: bool = False):
        ai_picker = cls(gpu=gpu)
        with(open(file_name, 'rb')) as f_in:
            picker = pickle.load(f_in)
            ai_picker.model = picker.model
            ai_picker.model.to(ai_picker.device)
            ai_picker.model = ai_picker.model.double()
            ai_picker.model = ai_picker.model.eval()
        # picker = cls()
        # picker.model.load_state_dict(torch.load(filename))
        # picker.model.eval()
        # picker.model.to(self.device)
        #
        # picker.in_channels = in_channels
        # self.base_filters = base_filters
        # self.kernel_size = kernel_size
        # self.stride = stride
        # self.n_classes = n_classes
        # self.groups = groups
        # self.n_block = n_block
        # self.gpu = gpu
        # self.n_sample = n_sample
        # self.sampling_rate = sampling_rate

        # if device == 'gpu':
        #     device = torch.device("cuda:0" if
        #                           torch.cuda.is_available() else "cpu")
        #     if device == "cpu":
        #         logger.warning('GPU is not available, '
        #                        'the CPU will be used')
        #     else:
        #         logger.info('GPU will be used')
        # else:
        #     device = 'cpu'
        #     logger.info('The CPU will be used')

        return ai_picker


def read_picker(file_name):
    return AIPicker.read(file_name)






from torch.utils.data import Dataset
from pathlib import Path
from uquake import read, read_events
from uquake.core import UTCDateTime
from uquake.core.logging import logger
from uquake.core.stream import Stream
from uquake.core.event import Event, Catalog
import numpy as np
from obspy.signal.trigger import (recursive_sta_lta, plot_trigger, coincidence_trigger,
                                  trigger_onset)
import matplotlib.pyplot as plt
from useis.ai.model import generate_spectrogram
from PIL import Image, ImageOps
from useis.ai.database import Record
from event_type_lookup import event_type_lookup
from scipy.stats import skew, kurtosis

from useis.processors import classifier
from importlib import reload
reload(classifier)
# from useis.processors.classifier import Classifier
from ipdb import set_trace
import multiprocessing
from tqdm import tqdm

np.random.seed(124)

##
# Frame size

raw_data_path = Path('/data_1/ot-reprocessed-data')
classifier_project = classifier.Classifier('/data_1/projects/', 'classifier', 'OT',
                                           reset_training=False)
training_data_path = classifier_project.paths.training_dataset
# training_data_path.mkdir(parents=True, exist_ok=True)

context_trace = raw_data_path.glob('*.context_mseed')

window_length_seconds = np.floor(np.logspace(0, 1, 3))
sampling_rates = np.array([6000])

frame_size = 1  # frame size in second HNAS request data every second from the Paladin


def test_impulsiveness(trace):

    # Load the signal from a file or generate it using numpy
    signal = trace.detrend('demean').detrend('linear').data
    signal = signal / np.max(np.abs(signal))

    # Compute the signal statistics
    std = np.std(signal)
    mad = np.median(np.abs(signal - np.median(signal)))
    # skewness = skew(signal)
    # kurt = kurtosis(signal)
    # print(mad, skewness, kurt)

    # Classify the signal based on its statistics
    if mad/std > 1e-3:
        return False
    else:
        return True


def window_signal(stream: Stream, trigger_on: float = 0.5, trigger_off: float = 0.1,
                  sta: float = 0.5, lta: float = 2):
    stream = stream.composite()
    for i, tr in enumerate(stream):
        stream[i].data /= np.std(stream[0].data)
    df = stream[0].stats.sampling_rate
    data = np.hstack([stream[0].data[-1::-1], stream[0].data])
    cft = recursive_sta_lta(data, int(sta * df), int(lta * df))
    cft = cft[len(cft)-len(stream[0].data):]
    triggers = trigger_onset(cft / np.nanmax(cft), trigger_on, trigger_off)

    start_time = stream[0].stats.starttime + triggers[0][0] / df
    end_time = stream[0].stats.starttime + triggers[-1][1] / df

    return start_time, end_time


# def create_spectrograms(stream: Stream, event_type: str, end_time: UTCDateTime,
#                         magnitude: float, expect_impulsive=True):
#
#     init_sampling_rate = stream[0].stats.sampling_rate
#     # from ipdb import set_trace
#     # set_trace()
#     starttime = stream[0].stats.starttime
#     end_time_sample = (end_time - starttime) * init_sampling_rate
#     for sampling_rate in sampling_rates:
#         end_time_resampled = starttime + end_time_sample / sampling_rate
#         for window_length in window_length_seconds:
#             trs_resampled = []
#             for tr in stream.copy():
#                 # if expect_impulsive:
#                 #     if not test_impulsiveness(tr):
#                 #         continue
#                 tr_resampled = tr.copy()
#                 tr_resampled.stats.sampling_rate = sampling_rate
#                 trs_resampled.append(tr_resampled)
#
#             if not trs_resampled:
#                 continue
#
#             st_resampled = Stream(traces=trs_resampled)
#             # st_resampled = stream.copy().resample(sampling_rate)
#             start_time = end_time_resampled - window_length
#             st_trimmed = st_resampled.copy().trim(starttime=start_time,
#                                                   endtime=end_time_resampled,
#                                                   pad=True, fill_value=0)
#             st_trimmed = st_trimmed.taper(max_percentage=0.01, max_length=1e-3)
#
#             # from ipdb import set_trace
#             # set_trace()
#
#             spectrograms = generate_spectrogram(st_trimmed)
#
#             filenames = write_spectrogram(spectrograms, event_type, end_time_resampled,
#                                           magnitude, window_length)
#
#             for filename, tr in zip(filenames, st_trimmed):
#                 plt.clf()
#                 plt.plot(tr.data, 'k')
#                 plt.axis('off')
#                 plt.tight_layout()
#                 filename = Path(filename)
#                 classifier_dataset_path = classifier_project.paths.training_dataset
#                 plt.savefig(f'{classifier_dataset_path / event_type / filename.stem}'
#                             f'_td.png')


# def write_spectrogram(spectrograms, event_type, end_time, magnitude, duration):
#
#     filenames = []
#     for i, spectrogram in enumerate(spectrograms):
#         if np.mean(spectrogram) == 0:
#             continue
#
#         filename = f'{event_type}_{end_time}_mag{magnitude:0.2f}_{duration}sec_ch{i}.png'
#         data_path = training_data_path / event_type
#         data_path.mkdir(parents=True, exist_ok=True)
#         path = training_data_path / event_type / filename
#         spectrogram.save(path, format='png')
#         filenames.append(filename)
#
#     return filenames


def create_blast_training_set(stream: Stream, event: Event, filename: str):
    # find the start and end of the blast train using sta/lta

    event_id = filename.name

    blast_start_time, blast_end_time = window_signal(stream)

    end_time = blast_start_time + 0.5
    ct = 1

    original_event_type = event.event_type

    while end_time <= blast_end_time:
        classifier_project.create_spectrogram_training(stream, 'blast',
                                                       original_event_type,
                                                       event_id, end_time,
                                                       event.magnitudes[-1].mag)
        end_time += 1


def find_end_signal(stream: Stream, event: Event):
    # finding the start and end of the signal
    p_pick_time = None
    for arrival in event.preferred_origin().arrivals:
        pick = arrival.get_pick()
        if pick.site == stream[0].stats.site:
            if arrival.phase.upper() == 'P':
                p_pick_time = pick.time

    if p_pick_time is None:
        tt_grid = classifier_project.travel_times.select(
            seed_labels=[stream[0].stats.site], phase='P')[0]
        p_pick_time = event.preferred_origin().time + \
                      tt_grid.interpolate(event.preferred_origin().loc)[0]

    # estimating the event duration
    # measuring the background energy

    # background = stream.copy().trim(starttime=p_pick_time-0.5, endtime=p_pick_time-0.1)
    # background_energy = np.mean(background[0].data ** 2)
    #
    # signal = stream.copy().trim(starttime=p_pick_time).detrend('demean').detrend(
    #     'linear')[0].data
    #
    # energy_window = 100  # sample
    # for i in range(len(signal) - energy_window):
    #     energy_remaining = np.mean(signal[i:i+energy_window] ** 2)
    #     if energy_remaining < 10 * background_energy:
    #         end_signal_t = p_pick_time + i / stream[0].stats.sampling_rate
    #         break

    start_signal = p_pick_time
    end_signal = None

    return start_signal, end_signal


def create_seismic_event_training_set(stream: Stream, event: Event, filename: str):

    start_signal, end_signal = find_end_signal(stream, event)
    # duration = end_signal - start_signal
    # min_start_window = start_signal + 0.3 * duration - np.min(window_length_seconds)
    # max_end_window = end_signal - 0.3 * duration + np.min(window_length_seconds)

    event_type = event_type_lookup[event.event_type]

    window_range = 0.6
    window_offset = 0.2

    perturbations = np.random.rand(5) * window_range + window_offset
    end_signals = [start_signal + t for t in perturbations]
    event_id = filename.name

    original_event_type = event.event_type

    for end_signal in end_signals:
        if event.preferred_magnitude():
            magnitude = event.preferred_magnitude().mag
        else:
            magnitude = event.magnitudes[-1].mag
        classifier_project.create_spectrogram_training(stream.copy(), event_type,
                                                       original_event_type,
                                                       event_id, end_signal,
                                                       magnitude,
                                                       simulate_magnitude=True)


def create_noise_training_set(stream: Stream, event: Event, filename: Path):
    end_signal = stream[0].stats.starttime + 12
    event_id = filename.name
    original_event_type = event.event_type
    classifier_project.create_spectrogram_training(stream, 'noise', original_event_type,
                                                   event_id, end_signal,
                                                   event.magnitudes[-1].mag,
                                                   expect_impulsive=False)


def event_size_scaling(stream: Stream, event: Event, end_signal, output_number=3,
                       magnitude_range=[-1, 3], magnitude_resolution=0.01, bvalue=1,
                       scaling_factor=2):
    """
    The function transform the stream to simulate larger magnitude.
    The scaling is performed by altering the frequency content of the waveform.
    The corner frequency is reduced by a factor 3 for every unit of magnitude.
    Some reference on why a factor of 3 is used can be found in
    https://agupubs.onlinelibrary.wiley.com/doi/full/10.1029/2004JB003084

    other authors have suggested a scaling to be closer to 2.

    Note that the distribution of magnitude produced in uniform. This is to counter-
    balance the typical distribution that is skewed towards smaller magnitude.

    @param stream: the waveform
    @type stream: uquake.core.stream.Stream
    @param event: the event information
    @type event: uquake.core.event.Event
    @param output_number: number of output.
    @param magnitude_range: the range of magnitude to randomly select from
    @param magnitude_resolution: the magnitude resolution
    @param bvalue: The b-value of the natural distribution to counterbalance
    @param scaling_factor: determine how the frequency scales with the increase or
    decrease in magnitude.
    """
    try:
       magnitude = event.preferred_magnitude().mag
    except Exception as e:
        logger.warning(e)
        magnitude = event.magnitudes[-1].mag
    if magnitude == -999:
        return
    streams = []
    mags = []
    end_signals = []
    for i in range(output_number):
        target_mag = generate_random_magnitude(range=magnitude_range,
                                               resolution=magnitude_resolution,
                                               bvalue=bvalue)
        mags.append(target_mag)
        frequency_scaling = scaling_factor ** (target_mag - magnitude)
        trs = []
        for tr in stream.copy():
            tmp = (end_signal - tr.stats.starttime) * tr.stats.sampling_rate
            end_signal = tr.stats.starttime + tmp / frequency_scaling
            tr.stats.sampling_rate /= frequency_scaling
            trs.append(tr)
            end_signals.append(end_signal)
        streams.append(Stream(traces=trs))
    return {'streams': streams, 'magnitudes': mags, 'end_signals': end_signals}


def generate_random_magnitude(range=[-2, 2.3]):
    r = np.max(range) - np.min(range)
    return np.random.rand(1)[0] * r + np.min(range)


durations = []
magnitudes = []


def create_training_dataset(context_trace):
    try:

        event_id = context_trace.name
        if classifier_project.event_exists(event_id):
            logger.info('event already exists in the database')
            return

        seismogram_file = context_trace

        try:
            st = read(context_trace).detrend('demean').detrend('linear')
        except Exception as e:
            logger.error(e)
            return
        event = read_events(context_trace.with_suffix('.xml'))[0]
        if np.sum(st.composite()[0].data) / len(st.composite()[0].data) > 0.5:
            return

        if event_type_lookup[event.event_type] == 'blast':
            create_blast_training_set(st, event, seismogram_file)

        elif event_type_lookup[event.event_type] == 'seismic event':

            create_seismic_event_training_set(st.copy(), event, seismogram_file)

        elif 'noise' in event_type_lookup[event.event_type]:
            create_noise_training_set(st, event, seismogram_file)
    except Exception as e:
        logger.error(e)


if __name__ == "__main__":
    mseed_files = [f for f in raw_data_path.glob('*.context_mseed')]

    # num_cores = multiprocessing.cpu_count()
    # pool = multiprocessing.Pool(10)

    # list(tqdm(pool.imap(create_training_dataset, mseed_files),
    #           total=len(mseed_files)))

    for context_trace in tqdm(mseed_files):
        # try:
        create_training_dataset(context_trace)
        # except Exception as e:
        #     logger.error(e)
        # # print(context_trace)





# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['useis',
 'useis.ai',
 'useis.clients',
 'useis.core',
 'useis.examples.classifier',
 'useis.examples.picker',
 'useis.processors',
 'useis.sandbox',
 'useis.sandbox.ai_picker',
 'useis.sandbox.demo_kafka',
 'useis.sandbox.event_classification',
 'useis.sandbox.event_classification.inventory',
 'useis.sandbox.tomography',
 'useis.scripts',
 'useis.scripts.ai',
 'useis.services',
 'useis.services.file_server',
 'useis.services.grid',
 'useis.services.models',
 'useis.settings',
 'useis.tomography']

package_data = \
{'': ['*']}

install_requires = \
['dynaconf>=3.1.4,<4.0.0',
 'fastapi>=0.68.1,<0.69.0',
 'furl>=2.1.2,<3.0.0',
 'myst-parser>=0.15.1,<0.16.0',
 'numpy>=1.3.0,<2.0.0',
 'pillow>=9.4.0,<10.0.0',
 'pydantic>=1.8.2,<2.0.0',
 'pyproj==3.4.1',
 'python-multipart>=0.0.5,<0.0.6',
 'rinohtype>=0.5.3,<0.6.0',
 'scikit-learn>=1.0.1,<2.0.0',
 'uquake>=1.4.5,<2.0.0',
 'uvicorn>=0.15.0,<0.16.0']

extras_require = \
{':extra == "docs"': ['Sphinx>=4.1.2,<5.0.0', 'sphinx-rtd-theme>=0.5.2,<0.6.0']}

entry_points = \
{'console_scripts': ['grid_service = useis.services.grid.server:start']}

setup_kwargs = {
    'name': 'useis',
    'version': '1.1.5',
    'description': '',
    'long_description': 'None',
    'author': 'jpmercier',
    'author_email': 'jpmercier01@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



from useis.processors import tomography
from importlib import reload

reload(tomography)
tomo = tomography.Tomography('/data_1/projects/tomo_test', 'tomo_test', 'TGT')

# tomo.synthetic()

tomo.__add_random_events__(1000)
tomo.__add_random_travel_times__()

rays = tomo.ray_tracing()
ray = rays[0]


# from uquake.grid.nlloc import VelocityGrid3D
# import numpy as np
# import scipy as sp
# import matplotlib.pyplot as plt
#
# # Grid parameters
# network_code = 'test'
# spacings = [10, 10, 10]
# origin = [0, 0, 0]
# dims = [100, 100, 100]
#
# nsensors = 12
# nevents = 1000
#
# # Building the velocity model
# velocity = VelocityGrid3D(network_code, dims, origin, spacings)
# velocity.fill_random(5000, 200, 5)
#
# # generate random events in the grid
# events = velocity.generate_random_points_in_grid(nevents)
# event_grid = velocity.transform_to_grid(events)
#
# # generate random sensors in the grid
# sensors = velocity.generate_random_points_in_grid(nsensors)
# sensor_grid = velocity.transform_to_grid(sensors)
# sensor_labels = [f'sta{i}' for i in range(0, nsensors)]
#
# times = velocity.to_time_multi_threaded(sensors, sensor_labels)
#
#
# ray = times[0].ray_tracer(events)
# nodes = velocity.transform_to_grid(ray.nodes)
#
# plt.clf()
# plt.imshow(velocity.data[:,:,10].T)
# plt.colorbar()
#
# plt.plot(event_grid[:, 0], event_grid[:, 1], 'k.')
# plt.plot(sensor_grid[:, 0], sensor_grid[:, 1], 'rv')
# plt.plot(nodes[:,0], nodes[:,1], 'r')
#
# plt.show()





from uquake.waveform.amp_measures import calc_velocity_flux
from uquake.waveform.mag import calculate_energy_from_flux

# from uquake.processors.processing_unit import ProcessingUnit

# from uquake.core.focal_mechanism import calc

from uquake.waveform.simple_mag import moment_magnitude
from uquake.waveform.amp_measures import measure_pick_amps
from ..core.project_manager import ProjectManager
from uquake.core.stream import Stream
from uquake.core.event import Catalog


class Magnitude(ProjectManager):

    def __init__(self, base_projects_path: Path, project_name: str,
                 network_code: str, use_srces: bool = False):
        """
        Object to control NLLoc execution and manage the required grids, inputs
        and outputs
        :param base_projects_path:
        :param project_name:
        :param network_code:
        :param use_srces:
        """

        super().__init__(base_projects_path, project_name, network_code,
                         use_srces=use_srces)

    def calculate_magnitude(self, stream: Stream, cat: Catalog):

        return moment_magnitude(stream, cat, self.inventory,
                                self.p_velocity, self.s_velocity)


    #
    #
    #
    # def process(
    #     self,
    #     **kwargs
    # ):
    #
    #     stream = kwargs['stream']
    #     cat = kwargs['cat']
    #
    #     inventory = self.settings.inventory
    #     stream.attach_response(inventory)
    #
    #     vp, vs = velocity.get_velocities()
    #     cat_moment = moment_magnitude(stream, cat, inventory, vp, vs)
    #
    #     cat_flux = calc_velocity_flux(stream, cat_moment, inventory,
    #                                   phase_list=['P', 'S'])
    #     cat_energy = calculate_energy_from_flux(cat_flux, inventory, vp, vs,
    #                                             use_sdr_rad=False)
    #
    #     mag = cat_energy[0].magnitudes[-1]
    #     energy_p = mag.energy_p_joule
    #     energy_s = mag.energy_s_joule
    #     energy = mag.energy_joule
    #     cat_energy[0].preferred_magnitude().energy_p_joule = energy_p
    #     cat_energy[0].preferred_magnitude().energy_s_joule = energy_s
    #     cat_energy[0].preferred_magnitude().energy_joule = energy
    #
    #     return cat_energy




from confluent_kafka import Consumer
import sys


conf = {'bootstrap.servers': "localhost:29092",
        'group.id': 'test',
        'auto.offset.reset': 'smallest'}

c = Consumer(conf)

running = True


def basic_consume_loop(consumer, topics):
    try:
        consumer.subscribe(topics)

        while running:
            msg = consumer.poll(timeout=1.0)
            if msg is None: continue

            if msg.error():
                print('error')
                # if msg.error().code() == #KafkaError._PARTITION_EOF:
                #     # End of partition event
                #     sys.stderr.write('%% %s [%d] reached end at offset %d\n' %
                #                      (msg.topic(), msg.partition(), msg.offset()))
                # elif msg.error():
                #     print(msg.error())
                #     # raise KafkaException(msg.error())
            else:
                print(msg)
    finally:
        # Close down consumer to commit final offsets.
        consumer.close()


def shutdown():
    running = False


basic_consume_loop(c, ['test'])



from pydantic import BaseModel
from typing import Optional, List


class ClassifierResults(BaseModel):
    networks: List[str]
    stations: List[str]
    channels: List[str]
    raw_outputs: List[List[float]]
    predicted_classes: List[str]
    probabilities: List[List[float]]

    class Config:
        schema_extra = {
            "example": {
                "raw_outputs": [
                    [1.0, 2.0, 3.0],
                    [4.0, 5.0, 6.0],
                    [7.0, 8.0, 9.0]
                ],
                "labels": ["class1", "class2", "class3"],
                "probabilities": [
                    [0.5, 0.3, 0.2],
                    [0.2, 0.6, 0.2],
                    [0.1, 0.3, 0.6]
                ]
            }
        }



import pyproj
from pydantic import BaseModel, validator, conlist


class Latitude(BaseModel):
    latitude: float

    @validator('latitude')
    def range(cls, v):
        assert -90 <= v <= 90, 'Latitude must be between -180 and 180'
        return v


class Longitude(BaseModel):
    longitude: float

    @validator('longitude')
    def range(cls, v):
        assert -180 <= v <= 180, 'Latitude must be between -90 and 90'
        return v


class Projection(object):
    def __init__(self, global_epsg_code: int = 4326,
                 local_epsg_code: int = 32613,
                 offset: conlist(float, min_items=3, max_items=3) = [0, 0, 0]):
        self.global_epsg_code = global_epsg_code
        self.local_epsg_code = local_epsg_code
        self.offset = offset

    def __repr__(self):
        return f"""[projection]
global.epsg = {self.global_epsg_code}
local.epsg = {self.local_epsg_code}
local.offset = {self.offset} 
"""

    def transform_to_local(self, latitude: float = 0, longitude: float = 0,
                           z: float = 0):
        Latitude(latitude=latitude)
        Longitude(longitude=longitude)
        transformer = pyproj.Transformer.from_crs(self.global_epsg_code,
                                                  self.local_epsg_code,
                                                  always_xy=True)
        easting, northing = transformer.transform(longitude, latitude)

        easting += self.offset[0]
        northing += self.offset[1]
        z += self.offset[2]

        return easting, northing, z

    def transform_to_global(self, easting: float = 0, northing: float = 0,
                            z: float = 0):
        transformer = pyproj.Transformer.from_crs(self.local_epsg_code,
                                                  self.global_epsg_code,
                                                  always_xy=True)

        easting -= self.offset[0]
        northing -= self.offset[1]
        z -= self.offset[2]

        longitude, latitude = transformer.transform(easting, northing)

        return longitude, latitude, z

    def write(self, filename):
        with open(filename, 'w') as fout:
            fout.write(str(self))


class UnitConversion(object):
    pass



from ..core.project_manager import ProjectManager


class FileManager(ProjectManager):
    def __init__(self, base_projects_path, project_name, network_code, waveform_path='',
                 quakeml_path='', init_catalogue=True, **kwargs):

        super().__init__(base_projects_path, project_name, network_code)

        self.paths['waveforms'] = waveform_path
        self.paths['quakeml'] = quakeml_path

        self.paths['catalog'] = self.base_projects_path / catalog





        # self.files['catalog_index'] = self.paths.
